"""
Constants we use throughout the app
"""

# Django
from django.utils.text import slugify

# AA Sovereignty Timer
from sovtimer import __version__

VERBOSE_NAME = "Sovereignty Timer for Alliance Auth"

verbose_name_slugified: str = slugify(VERBOSE_NAME, allow_unicode=True)
github_url: str = "https://github.com/ppfeufer/aa-sov-timer"

USER_AGENT = f"{verbose_name_slugified} v{__version__} {github_url}"



"""
pages url config
"""

# Django
from django.urls import path

# AA Sovereignty Timer
from sovtimer import views

app_name: str = "sovtimer"

urlpatterns = [
    path("", views.dashboard, name="dashboard"),
    # ajax call
    path("dashboard_data/", views.dashboard_data, name="dashboard_data"),
]



"""
the tasks
"""

# Third Party
from bravado.exception import HTTPBadGateway, HTTPGatewayTimeout, HTTPServiceUnavailable
from celery import shared_task

# Django
from django.core.cache import cache
from django.db import transaction

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.core.esitools import is_esi_online
from eveuniverse.models import EveEntity, EveSolarSystem

# AA Sovereignty Timer
from sovtimer import __title__
from sovtimer.models import Campaign, SovereigntyStructure

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

DEFAULT_TASK_PRIORITY = 6
ESI_ERROR_LIMIT = 50
ESI_TIMEOUT_ONCE_ERROR_LIMIT_REACHED = 60
ESI_SOV_STRUCTURES_CACHE_KEY = "sov_structures_cache"

# params for all tasks
TASK_DEFAULT_KWARGS = {
    "time_limit": 1200,  # stop after 20 minutes
}

# params for tasks that make ESI calls
TASK_ESI_KWARGS = {
    **TASK_DEFAULT_KWARGS,
    **{
        "autoretry_for": (
            OSError,
            HTTPBadGateway,
            HTTPGatewayTimeout,
            HTTPServiceUnavailable,
        ),
        "retry_kwargs": {"max_retries": 3},
        "retry_backoff": True,
    },
}


@shared_task(**TASK_DEFAULT_KWARGS)
def run_sov_campaign_updates() -> None:
    """
    update all sov campaigns
    """

    if not is_esi_online():
        logger.info(
            "ESI is currently offline. Can not start ESI related tasks. Aborting"
        )
        return

    logger.info("Updating sovereignty structures and campaigns from ESI.")

    update_sov_structures.apply_async(priority=DEFAULT_TASK_PRIORITY)
    update_sov_campaigns.apply_async(priority=DEFAULT_TASK_PRIORITY)


@shared_task(**{**TASK_ESI_KWARGS}, **{"base": QueueOnce})
def update_sov_campaigns() -> None:
    """
    update campaigns
    """

    campaigns_from_esi = Campaign.sov_campaigns_from_esi()

    if campaigns_from_esi:
        with transaction.atomic():
            campaigns = []

            campaign_count = 0

            for campaign in campaigns_from_esi:
                EveEntity.objects.get_or_create(id=campaign["defender_id"])

                campaign_current__defender_score = campaign["defender_score"]

                try:
                    campaign_previous = Campaign.objects.get(
                        campaign_id=campaign["campaign_id"]
                    )

                    campaign_previous__progress_previous = (
                        campaign_previous.progress_previous
                    )

                    campaign_previous__progress = campaign_previous.defender_score
                    campaign_current__progress_previous = campaign_previous__progress

                    if campaign_previous__progress == campaign_current__defender_score:
                        campaign_current__progress_previous = (
                            campaign_previous__progress_previous
                        )
                except Campaign.DoesNotExist:
                    campaign_current__progress_previous = (
                        campaign_current__defender_score
                    )

                campaigns.append(
                    Campaign(
                        attackers_score=campaign["attackers_score"],
                        campaign_id=campaign["campaign_id"],
                        defender_score=campaign["defender_score"],
                        event_type=campaign["event_type"],
                        start_time=campaign["start_time"],
                        structure_id=campaign["structure_id"],
                        progress_current=campaign_current__defender_score,
                        progress_previous=campaign_current__progress_previous,
                    )
                )

                campaign_count += 1

            Campaign.objects.all().delete()
            Campaign.objects.bulk_create(
                campaigns,
                batch_size=500,
                ignore_conflicts=True,
            )

            EveEntity.objects.bulk_update_new_esi()

            logger.info(f"{campaign_count} sovereignty campaigns updated from ESI.")


@shared_task(**{**TASK_ESI_KWARGS}, **{"base": QueueOnce})
def update_sov_structures() -> None:
    """
    update structures
    """

    if cache.get(ESI_SOV_STRUCTURES_CACHE_KEY) is None:
        logger.debug("UPDATING SOV STRUCTURES")
        structures_from_esi = SovereigntyStructure.sov_structures_from_esi()

        if structures_from_esi:
            with transaction.atomic():
                esi_structure_ids = []
                sov_structures = []

                structure_count = 0

                for structure in structures_from_esi:
                    esi_structure_ids.append(structure["structure_id"])

                    structure_alliance, _ = EveEntity.objects.get_or_create(
                        id=structure["alliance_id"]
                    )

                    structure_solar_system = EveSolarSystem.objects.get(
                        id=structure["solar_system_id"]
                    )

                    vulnerability_occupancy_level = 1
                    if structure["vulnerability_occupancy_level"]:
                        vulnerability_occupancy_level = structure[
                            "vulnerability_occupancy_level"
                        ]

                    vulnerable_end_time = None
                    if structure["vulnerable_end_time"]:
                        vulnerable_end_time = structure["vulnerable_end_time"]

                    vulnerable_start_time = None
                    if structure["vulnerable_start_time"]:
                        vulnerable_start_time = structure["vulnerable_start_time"]

                    sov_structures.append(
                        SovereigntyStructure(
                            alliance=structure_alliance,
                            solar_system=structure_solar_system,
                            structure_id=structure["structure_id"],
                            structure_type_id=structure["structure_type_id"],
                            vulnerability_occupancy_level=vulnerability_occupancy_level,
                            vulnerable_end_time=vulnerable_end_time,
                            vulnerable_start_time=vulnerable_start_time,
                        )
                    )

                    structure_count += 1

                SovereigntyStructure.objects.bulk_create(
                    sov_structures,
                    batch_size=500,
                    ignore_conflicts=True,
                )

                EveEntity.objects.bulk_update_new_esi()
                SovereigntyStructure.objects.exclude(pk__in=esi_structure_ids).delete()

                cache.set(ESI_SOV_STRUCTURES_CACHE_KEY, True, 120)

                logger.info(
                    f"{structure_count} sovereignty structures updated from ESI."
                )



"""
The views
"""

# Standard Library
import datetime as dt

# Django
from django.contrib.auth.decorators import login_required, permission_required
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse, JsonResponse
from django.shortcuts import render
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.evelinks.eveimageserver import alliance_logo_url
from allianceauth.eveonline.templatetags.evelinks import (
    dotlan_alliance_url,
    dotlan_region_url,
)
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Sovereignty Timer
from sovtimer import __title__
from sovtimer.models import Campaign, SovereigntyStructure

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("sovtimer.basic_access")
def dashboard(request: WSGIRequest) -> HttpResponse:
    """
    Index view
    """

    logger.info("Module called by %s", request.user)

    context = {}

    return render(request, "sovtimer/dashboard.html", context)


def dashboard_data(request: WSGIRequest) -> JsonResponse:
    """
    ajax call
    get dashboard data
    :param request:
    :return:
    """

    data = []

    sovereignty_campaigns = Campaign.objects.select_related(
        "structure",
        "structure__alliance",
        "structure__solar_system",
        "structure__solar_system__eve_constellation",
        "structure__solar_system__eve_constellation__eve_region",
    ).filter(structure__isnull=False)

    campaign_systems = sovereignty_campaigns.values_list(
        "structure__solar_system_id", flat=True
    )

    sovereignty_structures = SovereigntyStructure.objects.filter(
        solar_system_id__in=set(campaign_systems)
    )

    if sovereignty_campaigns:
        for campaign in sovereignty_campaigns:
            # defender name
            defender_name = campaign.structure.alliance.name
            defender_url = dotlan_alliance_url(campaign.structure.alliance)
            defender_logo_url = alliance_logo_url(
                alliance_id=campaign.structure.alliance.id, size=32
            )

            defender_name_html = (
                f'<a href="{defender_url}" target="_blank" rel="noopener noreferer">'
                f'<img class="aa-sovtimer-entity-logo-left" src="{defender_logo_url}" '
                f'alt="{defender_name}">{defender_name}</a>'
            )

            # solar system
            solar_system_name = campaign.structure.solar_system.name
            region_url = dotlan_region_url(
                campaign.structure.solar_system.eve_constellation.eve_region
            )
            campaign_system_name = campaign.structure.solar_system.name
            solar_system_url = f"{region_url}/{campaign_system_name}"
            solar_system_name_html = (
                f'<a href="{solar_system_url}" target="_blank" '
                f'rel="noopener noreferer">{campaign_system_name}</a>'
            )

            # constellation
            constellation_name = campaign.structure.solar_system.eve_constellation.name
            constellation_name_html = (
                f'<a href="//evemaps.dotlan.net/search?q={constellation_name}" '
                f'target="_blank" rel="noopener noreferer">{constellation_name}</a>'
            )

            # region
            region_name = (
                campaign.structure.solar_system.eve_constellation.eve_region.name
            )
            region_url = dotlan_region_url(
                campaign.structure.solar_system.eve_constellation.eve_region
            )
            region_name_html = (
                f'<a href="{region_url}" target="_blank" '
                f'rel="noopener noreferer">{region_name}</a>'
            )

            # activity defense multiplier
            structure_adm = 1
            for structure in sovereignty_structures:
                if (
                    structure.solar_system_id == campaign.structure.solar_system_id
                ) and structure.vulnerability_occupancy_level:
                    structure_adm = structure.vulnerability_occupancy_level

            # start time
            start_time = campaign.start_time

            # remaining time field
            remaining_time_in_seconds = dt.timedelta(
                seconds=(start_time.timestamp() - dt.datetime.now().timestamp())
            ).total_seconds()

            # campaign progress field
            campaign_progress_previous = campaign.progress_previous
            campaign_progress_previous_percentage = campaign_progress_previous * 100
            campaign_progress_previous_percentage_formatted = (
                f"{campaign_progress_previous_percentage:.0f}%"
            )

            if campaign_progress_previous_percentage < 10:
                campaign_progress_previous_percentage_formatted = (
                    "0" + campaign_progress_previous_percentage_formatted
                )

            campaign_pogress_current = campaign.progress_current
            campaign_pogress_current_percentage = campaign_pogress_current * 100
            campaign_pogress_current_percentage_formatted = (
                f"{campaign_pogress_current_percentage:.0f}%"
            )
            if campaign_pogress_current_percentage < 10:
                campaign_pogress_current_percentage_formatted = (
                    "0" + campaign_pogress_current_percentage_formatted
                )

            campaign_progress_html = campaign_pogress_current_percentage_formatted

            active_campaign = _("No")
            if remaining_time_in_seconds < 0:
                active_campaign = _("Yes")

                progress_text = _("Neither side has made any progress yet")
                campaign_progress_icon = (
                    "<i "
                    'class="material-icons aa-sovtimer-trend aa-sovtimer-trend-flat" '
                    f'title="{progress_text}">trending_flat</i>'
                )

                if campaign_progress_previous < campaign_pogress_current:
                    defender_progress_text = _("Defenders making progress")
                    campaign_progress_icon = (
                        "<i "
                        'class="material-icons aa-sovtimer-trend aa-sovtimer-trend-up" '
                        f'title="{defender_progress_text}">trending_up</i>'
                    )

                if campaign_progress_previous > campaign_pogress_current:
                    attacker_progress_text = _("Attackers making progress")
                    campaign_progress_icon = (
                        "<i "
                        'class="material-icons aa-sovtimer-trend aa-sovtimer-trend-down" '
                        f'title="{attacker_progress_text}">trending_down</i>'
                    )

                constellation_id = campaign.structure.solar_system.eve_constellation.id
                zkb_href = f"https://zkillboard.com/constellation/{constellation_id}/"
                zkb_icon = '<img src="/static/sovtimer/images/zkillboard.png">'
                constellation_killboard_link = (
                    f'<a href="{zkb_href}" target="_blank" rel="noopener noreferer" '
                    f'class="aa-sov-timer-zkb-icon">{zkb_icon}</a>'
                )

                campaign_progress_html = (
                    campaign_progress_previous_percentage_formatted
                    + campaign_progress_icon
                    + campaign_pogress_current_percentage_formatted
                    + constellation_killboard_link
                )

            data.append(
                {
                    # type column
                    "event_type": Campaign.Type(campaign.event_type).label,
                    # system column + filter
                    "solar_system_name": solar_system_name,
                    "solar_system_name_html": solar_system_name_html,
                    # constellazion column + filter
                    "constellation_name": constellation_name,
                    "constellation_name_html": constellation_name_html,
                    # region column + filter
                    "region_name": region_name,
                    "region_name_html": region_name_html,
                    # defender column + filter
                    "defender_name": defender_name,
                    "defender_name_html": defender_name_html,
                    # adm column
                    "adm": structure_adm,
                    # start column
                    "start_time": start_time,
                    # remaining column
                    "remaining_time": "",
                    "remaining_time_in_seconds": remaining_time_in_seconds,
                    # progress column
                    "campaign_progress": campaign_progress_html,
                    # acive filter column (hidden)
                    "active_campaign": active_campaign,
                }
            )

    return JsonResponse(data, safe=False)



"""
Test ajax calls
"""

# Django
from django.test import RequestFactory, TestCase
from django.urls import reverse

# Alliance Auth
from allianceauth.eveonline.models import EveCharacter

# Alliance Auth (External Libs)
from app_utils.testing import add_character_to_user, create_user_from_evecharacter

# AA Sovereignty Timer
from sovtimer.tests.fixtures.load_allianceauth import load_allianceauth
from sovtimer.tests.fixtures.load_sovtimer import load_sovtimer
from sovtimer.views import dashboard_data


class TestAjaxCalls(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()

        cls.factory = RequestFactory()

        # given
        cls.character_1001 = EveCharacter.objects.get(character_id=1001)
        cls.character_1002 = EveCharacter.objects.get(character_id=1002)
        cls.character_1003 = EveCharacter.objects.get(character_id=1003)
        cls.character_1004 = EveCharacter.objects.get(character_id=1004)
        cls.character_1005 = EveCharacter.objects.get(character_id=1005)
        cls.character_1101 = EveCharacter.objects.get(character_id=1101)

        cls.user_with_basic_access, _ = create_user_from_evecharacter(
            cls.character_1001.character_id,
            permissions=["sovtimer.basic_access"],
        )

        add_character_to_user(cls.user_with_basic_access, cls.character_1101)

        cls.user_without_access, _ = create_user_from_evecharacter(
            cls.character_1002.character_id
        )

        add_character_to_user(cls.user_without_access, cls.character_1003)

    def test_ajax_dashboard_data_no_campaigns(self):
        # given
        self.client.force_login(self.user_with_basic_access)
        request = self.factory.get(reverse("sovtimer:dashboard"))
        request.user = self.user_with_basic_access

        # when
        result = dashboard_data(request=request)

        # then
        self.assertEqual(result.status_code, 200)
        self.assertJSONEqual(result.content, [])

    def test_ajax_dashboard_data_with_campaigns(self):
        self.maxDiff = None
        # given
        load_sovtimer()
        self.client.force_login(self.user_with_basic_access)
        request = self.factory.get(reverse("sovtimer:dashboard"))
        request.user = self.user_with_basic_access

        # when
        result = dashboard_data(request=request)

        # then
        self.assertEqual(result.status_code, 200)
        # self.assertJSONEqual(
        #     result.content,
        #     [
        #         {
        #             "event_type": "TCU",
        #             "solar_system_name": "0-O6XF",
        #             "solar_system_name_html": '<a href="http://evemaps.dotlan.net/map/Esoteria/0-O6XF" target="_blank" rel="noopener noreferer">0-O6XF</a>',
        #             "constellation_name": "Q-2BI6",
        #             "constellation_name_html": '<a href="//evemaps.dotlan.net/search?q=Q-2BI6" target="_blank" rel="noopener noreferer">Q-2BI6</a>',
        #             "region_name": "Esoteria",
        #             "region_name_html": '<a href="http://evemaps.dotlan.net/map/Esoteria" target="_blank" rel="noopener noreferer">Esoteria</a>',
        #             "defender_name": "Wayne Enterprises",
        #             "defender_name_html": '<a href="http://evemaps.dotlan.net/alliance/Wayne_Enterprises" target="_blank" rel="noopener noreferer"><img class="aa-sovtimer-entity-logo-left" src="https://images.evetech.net/alliances/3001/logo?size=32" alt="Wayne Enterprises">Wayne Enterprises</a>',
        #             "adm": 6.0,
        #             "start_time": "2021-10-14T11:38:37Z",
        #             "remaining_time": "",
        #             "remaining_time_in_seconds": 5734.00943,
        #             "campaign_progress": "60%",
        #             "active_campaign": "No",
        #         }
        #     ],
        # )
